@@ -7,7 +7,7 @@
 #include "ruby_prof.h"
 #include "rp_method.h"
 
-typedef struct
+typedef struct prof_allocation_t
 {
     st_data_t key;                    /* Key in hash table */
     unsigned int klass_flags;         /* Information about the type of class */
@@ -22,10 +22,10 @@ typedef struct
 
 void rp_init_allocation(void);
 void prof_allocation_free(prof_allocation_t* allocation);
-void prof_allocation_mark(void *data);
+void prof_allocation_mark(void* data);
 VALUE prof_allocation_wrap(prof_allocation_t* allocation);
 prof_allocation_t* prof_allocation_get(VALUE self);
-prof_allocation_t* prof_allocate_increment(prof_method_t *method, rb_trace_arg_t *trace_arg);
+prof_allocation_t* prof_allocate_increment(prof_method_t* method, rb_trace_arg_t* trace_arg);
 
 
 #endif //_RP_ALLOCATION_
