@@ -8,7 +8,7 @@
 
 extern VALUE mMeasure;
 
-typedef double (*get_measurement)(rb_trace_arg_t *trace_arg);
+typedef double (*get_measurement)(rb_trace_arg_t* trace_arg);
 
 typedef enum
 {
@@ -18,7 +18,7 @@ typedef enum
     MEASURE_MEMORY
 } prof_measure_mode_t;
 
-typedef struct
+typedef struct prof_measurer_t
 {
     get_measurement measure;
     prof_measure_mode_t mode;
@@ -36,13 +36,14 @@ typedef struct prof_measurement_t
     VALUE object;
 } prof_measurement_t;
 
-prof_measurer_t *prof_get_measurer(prof_measure_mode_t measure, bool track_allocations);
-double prof_measure(prof_measurer_t *measurer, rb_trace_arg_t* trace_arg);
+prof_measurer_t* prof_get_measurer(prof_measure_mode_t measure, bool track_allocations);
+double prof_measure(prof_measurer_t* measurer, rb_trace_arg_t* trace_arg);
 
-prof_measurement_t *prof_measurement_create(void);
-VALUE prof_measurement_wrap(prof_measurement_t *measurement);
+prof_measurement_t* prof_measurement_create(void);
+void prof_measurement_free(prof_measurement_t* measurement);
+VALUE prof_measurement_wrap(prof_measurement_t* measurement);
 prof_measurement_t* prof_get_measurement(VALUE self);
-void prof_measurement_mark(void *data);
+void prof_measurement_mark(void* data);
 
 void rp_init_measure(void);
 
