@@ -10,13 +10,12 @@
 
 extern VALUE cProfile;
 
-typedef struct
+typedef struct prof_profile_t
 {
     VALUE running;
     VALUE paused;
 
     prof_measurer_t* measurer;
-    VALUE threads;
 
     VALUE tracepoints;
 
@@ -27,6 +26,7 @@ typedef struct
     thread_data_t* last_thread_data;
     double measurement_at_pause_resume;
     bool allow_exceptions;
+    bool merge_fibers;
 } prof_profile_t;
 
 void rp_init_profile(void);
